@@ -2,16 +2,16 @@
  Copyright (C) 2019-present Travers Ching
  
  This program is free software: you can redistribute it and/or modify
- it under the terms of the GNU Affero General Public License as
- published by the Free Software Foundation, either version 3 of the
- License, or (at your option) any later version.
+ it under the terms of the GNU General Public License as published by
+ the Free Software Foundation, either version 3 of the License, or
+ (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- GNU Affero General Public License for more details.
+ GNU General Public License for more details.
  
- You should have received a copy of the GNU Affero General Public License
+ You should have received a copy of the GNU General Public License
  along with this program.  If not, see <https://www.gnu.org/licenses/>.
  
  You can contact the author at:
@@ -106,7 +106,7 @@ struct Compress_Thread_Context {
       while (blocks_written % nthreads != thread_id) {
         std::this_thread::yield();
       }
-      writeSizeToFile4(*myFile, zsize);
+      writeSize4(*myFile, zsize);
       myFile->write(zblocks[thread_id].data(), zsize);
       blocks_written += 1;
 
@@ -127,7 +127,7 @@ struct Compress_Thread_Context {
       while (blocks_written % nthreads != thread_id) {
         std::this_thread::yield();
       }
-      writeSizeToFile4(*myFile, zsize);
+      writeSize4(*myFile, zsize);
       myFile->write(zblocks[thread_id].data(), zsize);
       blocks_written += 1;
 
